<?php
/**
 * Authserver, an OAuth2-based single-signon authentication provider written in PHP.
 *
 * Copyright (C) 2017 Lars Vierbergen
 *
 * his program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
namespace vierbergenlars\AuthserverStatsBundle\EventListener;

use Symfony\Component\EventDispatcher\EventSubscriberInterface;
use vierbergenlars\AuthserverStatsBundle\Event\StatsEvent;
use Symfony\Bridge\Doctrine\RegistryInterface;
use Symfony\Component\Security\Http\SecurityEvents;
use Symfony\Component\Security\Http\Event\InteractiveLoginEvent;
use vierbergenlars\AuthserverStatsBundle\Entity\LoginEntry;

class LoginStatsListener implements EventSubscriberInterface
{

    /**
     *
     * @var RegistryInterface
     */
    private $registry;

    public static function getSubscribedEvents()
    {
        return [
            StatsEvent::class => [
                'getLoginStats'
            ],
            SecurityEvents::INTERACTIVE_LOGIN => [
                'onInteractiveLogin'
            ]
        ];
    }

    public function __construct(RegistryInterface $registry)
    {
        $this->registry = $registry;
    }

    public function onInteractiveLogin(InteractiveLoginEvent $event)
    {
        $user = $event->getAuthenticationToken()->getUser();

        $loginEntry = new LoginEntry($user);
        $em = $this->registry->getManagerForClass(LoginEntry::class);
        $em->persist($loginEntry);
        $em->flush($loginEntry);
    }

    public function getLoginStats(StatsEvent $event)
    {
        if ($event->isEnabled('login')) {

            $event->setMuninConfig('login', [
                'graph_title' => 'Authserver logins',
                'graph_vlabel' => 'Events/${graph_period}',
                'graph_category' => 'authserver',
                'graph_period' => 'minute',
                'total.label' => 'Interactive logins',
                'total.type' => 'COUNTER'
            ]);

            $queryBuilder = $this->registry->getRepository(LoginEntry::class)->createQueryBuilder('e');
            /* @var $queryBuilder \Doctrine\ORM\QueryBuilder */

            $event->addStatistic('login.total', $queryBuilder->select('count(e)')
                ->getQuery()
                ->getSingleScalarResult());
        }

        if ($event->isEnabled('login_aggregate')) {
            $event->setMuninConfig('login_aggregate', [
                'graph_title' => 'Authserver logins',
                'graph_vlabel' => 'Number of logins',
                'graph_category' => 'authserver',
                '5min.label' => 'Logins in past 5 minutes',
                '1hour.label' => 'Logins in past hour',
                '1day.label' => 'Logins in past day',
                '1week.label' => 'Logins in past week'
            ]);

            $queryBuilder = $this->registry->getRepository(LoginEntry::class)->createQueryBuilder('e');
            /* @var $queryBuilder \Doctrine\ORM\QueryBuilder */
            $queryBuilder->select('count(e)')->where('e.loginTime > :time');

            foreach ([
                '5min',
                '1hour',
                '1day',
                '1week'
            ] as $timeAgo) {
                $event->addStatistic('login_aggregate.' . $timeAgo, $queryBuilder->getQuery()
                    ->setParameter('time', new \DateTime($timeAgo . ' ago'))
                    ->getSingleScalarResult());
            }
        }
    }
}